/*
    Copyright 2010 Marten Richter

    This file is part of VOBD and VDR Nalustripper patch.

    VOBD is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    VOBD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with VOBD; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
//portions from demuxerts.cc from vomp www.loggytronic.com/vomp.php in function ParsePATPMT(

#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#define  MAXPADDINGSTRIPPERS 16

#define VERSION "0.0.0.3"



class cTSPaddingStripper {
public:
    cTSPaddingStripper(int mypid);
    ~cTSPaddingStripper();

    unsigned char *ProcessTSPacket(unsigned char*inputpacket);
    virtual bool ProcessPayload(unsigned char* payload, int size, bool payloadstart)=0;

    inline int Pid() {return pid;};

protected:
    int in_conti;
    int out_conti;
    int pid;
};



class cTSPaddingStrippers {
public:
    cTSPaddingStrippers();
    ~cTSPaddingStrippers();

    unsigned char *ProcessTSPacket(unsigned char*inputpacket);

    void AddStripper(cTSPaddingStripper *strip);
    void DetachStripper(cTSPaddingStripper *strip);

protected:
    cTSPaddingStripper *strippers[MAXPADDINGSTRIPPERS];
};

class cTSNaluPaddingStripper: public cTSPaddingStripper {
public:
    cTSNaluPaddingStripper(int mypid);
    ~cTSNaluPaddingStripper();

    bool ProcessPayload(unsigned char* payload, int size, bool payloadstart);
protected:
    bool nalustrip12;
    unsigned int pattern;
};




cTSPaddingStrippers::cTSPaddingStrippers()
{
    for (int i=0;i<MAXPADDINGSTRIPPERS;i++)
    {
        strippers[i]=NULL;
    }
}

cTSPaddingStrippers::~cTSPaddingStrippers()
{
    for (int i=0;i<MAXPADDINGSTRIPPERS;i++)
    {
        if (strippers[i]){
            delete strippers[i];
        }
    }
}

void cTSPaddingStrippers::AddStripper(cTSPaddingStripper *strip)
{
    for (int i=0;i<MAXPADDINGSTRIPPERS;i++)
    {
        if (!strippers[i]){
            strippers[i]=strip;
            return;
        }
    }
}

void cTSPaddingStrippers::DetachStripper(cTSPaddingStripper *strip)
{
    for (int i=0;i<MAXPADDINGSTRIPPERS;i++)
    {
        if (strippers[i]==strip){
            strippers[i]=NULL;
            return;
        }
    }
}


unsigned char *cTSPaddingStrippers::ProcessTSPacket(unsigned char*input)
{

    int pid=(input[1] &0x1f)<<8 | input[2];
    if (pid == 0) return input;
    if (pid == 0x1fff) return NULL; //padding
    if (!(input[3] &0x10)) return input; //without payload it must be important adaption field information
    for (int i=0;i<MAXPADDINGSTRIPPERS;i++)
    {
        if (strippers[i] && strippers[i]->Pid()==pid)
        {
            return strippers[i]->ProcessTSPacket(input);
        }
    } 
    return input;
}


cTSPaddingStripper::cTSPaddingStripper(int mypid)
{
    pid=mypid;
    in_conti=-1;
    out_conti=0;
}

cTSPaddingStripper::~cTSPaddingStripper()
{
    //goo;
}


unsigned char *cTSPaddingStripper::ProcessTSPacket(unsigned char*input)
{
    int conti=input[3]&0x0f;
    unsigned char sizeadapt=0;
    if (input[3] &0x20) {
        sizeadapt=input[4]+1;
    }
    int size_payload=184-sizeadapt;
    if (ProcessPayload(input+4+sizeadapt, size_payload,input[1]&0x40)) {
        in_conti=conti;
        return NULL;
    }
    if (conti==in_conti) out_conti--;
    if (out_conti<0) out_conti+=16;
    in_conti=conti;
    out_conti=(out_conti+1)%16;
    input[3]=(input[3] &0xf0) | out_conti;
    return input;
}

cTSNaluPaddingStripper::cTSNaluPaddingStripper(int mypid): cTSPaddingStripper(mypid)
{
    nalustrip12=false;
    pattern=0xFFFFFFFF;
}

cTSNaluPaddingStripper::~cTSNaluPaddingStripper()
{

}

bool cTSNaluPaddingStripper::ProcessPayload(unsigned char* payload, int size, bool startpayload)
{
    unsigned char *pos=payload;
    unsigned char *end=pos+size;
    unsigned char *new_pos=payload;
    if (startpayload) {
        nalustrip12=false;
        pattern=0xFFFFFFFF;
    }
    
    while (pos <end) {
        if (nalustrip12) {
            if (pattern==0x000001 || pattern==0x00000) {
                nalustrip12=false;
                pos-=3;
            }
        } else if (pattern==0x000001 ) {
           
             if (((*pos) &0x1f)==0x0c ) { //padding 
                if (!nalustrip12) {
                    new_pos-=3;
                 }
                nalustrip12=true;
                
             } else {
                 if ((*pos)>=0xe0 && *pos<=0xef ) 
                 {
                     if (startpayload) {
                         *new_pos=*pos;
                         pos++; new_pos++;
                         if (pos>=end) {return (pos!=new_pos);}
                         *pos=0; //was new_pos
                         pos++; new_pos++;
                         if (pos>=end) { return (pos!=new_pos);}
                         *pos=0; //was new_pos
                     } else {
                         printf("Problem in pes size should not happen!\n");
                     }
                 }
             }
        }   
        pattern= ((pattern<< 8) | *pos)&0x00FFFFFF;
        if (!nalustrip12) {
            new_pos++;
        } 
        pos++;
    }
    if ((pos-new_pos)==size) return true;
    else return false;
}

int pmtpid=-1;

inline int ParsePATPMT(unsigned char *input)
{
    int pid=(input[1] &0x1f)<<8 | input[2];
  //  if (pid!=0 && pid!=pmtpid) return -1;
    unsigned char *buf=input+4;
    unsigned char sizeadapt=0;
    if (input[2] &0x20) {
        sizeadapt=input[5]+1;
    }
    buf+=sizeadapt;
  //  printf("pid %d",pid);

    if (pid==0) { //PAT
        int new_pmtpid = ((buf[11]<< 8) | (buf[12]));
        new_pmtpid&= 0x1FFF;
        pmtpid=new_pmtpid;
        return -1;
    } else if (pid==pmtpid) { //PMT
        //portions from demuxerts.cc from vomp www.loggytronic.com/vomp.php
        
        int sectionlength = ((buf[2] << 8) & 0x0F ) | buf[3];
        int p = 13; //skip fixed part of pmt
        while ( p < sectionlength) {
            int streamtype = *(buf+p);
            p++;
            int foundpid = (*(buf+p)<< 8) | *(buf+p+1);
            p += 2; //skip ES Pid
            int eslength = ((*(buf+p) << 8) & 0x0F ) | *(buf+p+1);
            p += 2; //skip ES length
            foundpid = foundpid & 0x1FFF; //clear upper 3 bits
            if (streamtype==0x1b) return foundpid;
            p += eslength; //skip es descriptor
        }
    }
    return -1;

}





int main( int argc, const char* argv[] )
{
	FILE *in;
    FILE* out;
    printf("Nalustripper Version %s \n",VERSION);
    if (argc<3) {
        printf("Too few arguments");
        exit(-1);
    }
    in=fopen64(argv[1],"rb");
    printf("Input file: %s\n",argv[1]);
    if (in==NULL) {
        printf("Cannot open %s",argv[1]);
        exit(-1);
    }
    out=fopen64(argv[2],"wb");
     printf("Output file: %s\n",argv[2]);
    if (in==NULL) {
        printf("Cannot create %s",argv[2]);
        exit(-1);
    }
    cTSPaddingStrippers strippers;
    cTSPaddingStripper *ourstripper;
    unsigned char tspacket[188];
    tspacket[0]=0x47;
    int h264pid=-1;
    int packet_cnt=0;
    while (!feof(in))
    {
        unsigned char byte;
        if (fread(&byte,1,1,in)==1 && byte==0x47)
        {
            if (fread(tspacket+1,1,187,in)==187) 
            {
                int newh264pid=ParsePATPMT(tspacket);
                if (h264pid==-1 && newh264pid!=-1) {
                    h264pid=newh264pid;
                    printf("\nFound new h264pid %d!\n",h264pid);
                    ourstripper=new cTSNaluPaddingStripper(h264pid);
                    strippers.AddStripper(ourstripper);
                } else if (newh264pid!=-1 && newh264pid!=h264pid) {
                    printf("\nRemove old h264pid %d!\n",h264pid);
                    strippers.DetachStripper(ourstripper);
                    delete ourstripper;
                    h264pid=newh264pid;
                     printf("\nFound new h264pid %d!\n",h264pid);
                    ourstripper=new cTSNaluPaddingStripper(h264pid);
                    strippers.AddStripper(ourstripper);
                }

                unsigned char *resultpacket=strippers.ProcessTSPacket(tspacket);
                if (resultpacket) {
                    fwrite(resultpacket,1,188,out);
                }
            }
            packet_cnt++;
            if (packet_cnt%100000==0) {printf("*");fflush(stdout);}
        }
    }
    printf("\n");fflush(stdout);
    fclose(in);
    fclose(out);

}




